#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <time.h>
#include "db.h"
#include "config.h"
#include "funcs.h"

void free_db(struct DB_STR *db) {
	while (db != NULL) {
		struct DB_STR *ptr = db->next;

		if (db->filename != NULL)
			free(db->filename);

		if (db->desc != NULL)
			free(db->desc);

		if (db->tags != NULL) {
			for (size_t i = 0; i < db->size; i++)
				if (db->tags[i] != NULL)
					free(db->tags[i]);

			free(db->tags);
		}

		free(db);
		db = ptr;
	}
}

int append_tag(struct DB_STR *db, const char *tok, const size_t allocated) {
	if (allocated == 0) {
		db->tags = malloc(sizeof(char *) * (db->size + 1));
		if (db->tags == NULL)
			return 1;

		db->tags[0] = strdup(tok);
		if (db->tags[0] == NULL) {
			free(db->tags);
			return 1;
		}
	}

	else {
		db->tags[allocated] = strdup(tok);
		if (db->tags[allocated] == NULL)
			return 1;
	}

	return 0;
}

struct DB_STR *new_db_value(const struct CONFIG cfg, char *str) {
	struct DB_STR *db = malloc(sizeof(struct DB_STR));
	if (db == NULL)
		return NULL;

	/* Prepare string */
	memset(db, 0, sizeof(struct DB_STR));
	char *ptr = strrchr(str, '\n');
	if (ptr != NULL)
		*ptr = '\0';

	/* Filename */
	char *tok_rest = str;
	char *tok = strtok_r(tok_rest, "|", &tok_rest);
	if (tok != NULL) {
		db->filename = strdup(tok);
		if (db->filename == NULL)
			goto NEW_DB_VALUE_CLOSE;
	}

	else
		goto NEW_DB_VALUE_CLOSE;

	/* Tags */
	tok = strtok_r(tok_rest, "|", &tok_rest);
	if (tok != NULL) {
		db->size = 1;
		for (size_t i = 0; i < strlen(tok); i++)
			if (tok[i] == ',' && tok[i + 1])
				db->size++;

		size_t allocated = 0;

		char *tok2_rest = tok;
		char *tok2;
		while ((tok2 = strtok_r(tok2_rest, ",", &tok2_rest)) != NULL) {
			if (append_tag(db, tok2, allocated))
				goto NEW_DB_VALUE_CLOSE;

			allocated++;
		}
	}

	/* Desc */
	tok = strtok_r(tok_rest, "|", &tok_rest);
	if (tok != NULL)
		db->desc = strdup(tok);

	/* Stats */
	struct stat sb;
	if (stat(db->filename, &sb) < 0)
		fprintf(stderr, "%s: stat: %s: %s\n", cfg.prog_name, db->filename, strerror(errno));

	db->birthtime = sb.st_mtime;

	return db;

	NEW_DB_VALUE_CLOSE:
		free_db(db);
		return NULL;
}

struct DB_STR *read_db(const struct CONFIG cfg) {
	FILE *fp = file_open(cfg.prog_name, "db", "r");
	if (fp == NULL)
		return NULL;

	int ret = 1;

	char *line = NULL;
	ssize_t bytes = 0;
	size_t n = 0;

	struct DB_STR *db = NULL;
	struct DB_STR *i = NULL;

	while ((bytes = getline(&line, &n, fp)) > 0) {
		if (line[0] == '\0' || line[0] == '\n')
			continue;

		db = new_db_value(cfg, line);
		if (db == NULL)
			goto READ_DB_CLOSE;

		db->next = i;
		i = db;
	}

	ret = 0;

	READ_DB_CLOSE:
		if (ret) {
			free_db(db);
			fprintf(stderr, "%s: db: %s\n", cfg.prog_name, strerror(errno));
		}

		if (line)
			free(line);

		fclose(fp);

	return db;
}
